// Arduino-Code for an ECG with AD8232, 1.3 inch LCD-Module ST7789, SD-Card reader.
// For reason of performance ECG is either displayed on LCD or saved on SD-Card.
// Original Code by Peter Balch adapted for my needs.
// This code is public domain
// 2024 Bernd Gruber
 

#include <SdFat.h>
#include <SPI.h>
#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7789.h> // Hardware-specific library for ST7789

const int TFT_WIDTH = 240;   // Für Waveshare Debo 1.3 inch Display
const int TFT_HEIGHT = 240;

const int TFT_CS  = 10;   // Pin-Belegungen für Display, DIN(MOSI)=D11, CLK(SCK)=13 per default
const int TFT_CD  = 7;
const int TFT_RST = 8;
const int TFT_BL  = 9;   // PWM-Pin für Display-Helligkeit

const int BUTTON = 5;   // Umschalter zwischen Display und SD-Karte, VCC (an) = Display, GND (aus) = SD-Karte
const int SDCARD = 6;   // CS-Pin der SD-Karte
const int EKG_IN = A2;  // Output  AD8232
const int SamplePeriod = 5; // mSec

uint8_t Trace[TFT_WIDTH];
uint16_t DisplayRepeat = TFT_WIDTH;
uint8_t ignoreBeats = 0;
int ButtonState = 0;         // Initialisierung für die Schalterstellung BUTTON
int lastButtonState = 1;

File ekgDaten;
SdFat SD;

Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_CD, TFT_RST);


//-----------------------------------------------------------------------------
// WriteSdCard
//-----------------------------------------------------------------------------
void WriteSdCard (int ecg)
{
  if (!((ekgDaten.print (millis())) && (ekgDaten.print (" ")) && (ekgDaten.println ((ecg))))) {     // Daten schreiben und Puffer leeren
    analogWrite(TFT_BL, 80);
    tft.setCursor(4,20);
    tft.setTextSize(3);
    tft.print("Fehler beim Schreiben");
    ekgDaten.close();
//    while (1);    // beenden oder nochmal probieren
    delay(2000);
  }
}
//-----------------------------------------------------------------------------
// DrawGridVLine
//-----------------------------------------------------------------------------
void DrawGridVLine(int x, int y1, int y2)
{
  int y;
  
  if (y1 > y2)
    DrawGridVLine(x, y2, y1); else
  {
    if (x*5 % (1000/SamplePeriod) == 0)
      tft.drawFastVLine(x, y1, y2-y1+1, ST77XX_RED); else
    if (x*25 % (1000/SamplePeriod) == 0)
      tft.drawFastVLine(x, y1, y2-y1+1, 0xffaa00); else
    {
      tft.drawFastVLine(x, y1, y2-y1+1, ST77XX_BLACK);   
      for (y=y1; y <= y2; y++)
        if (y % 40 == 0)
          tft.drawPixel(x,y,ST77XX_RED); else
        if (y % 8 == 0)
          tft.drawPixel(x,y,0xffaa00); 
    }
  }
}
//-----------------------------------------------------------------------------
// calcBPM
//-----------------------------------------------------------------------------
void calcBPM(uint8_t ecg, uint16_t x)
{
  static uint8_t prevy = 0;
  uint32_t t;
  static uint32_t prevt = 0;
  static int InPeak = 0;
  static int bpm = 50;
  const int threshHi = 40;
  const int threshLo = threshHi - 5;
  int b;
  static bool first = true;

  if (ecg > prevy)
    prevy++; else
    prevy--;

  if (InPeak > 0) {
    InPeak++;
    if (ecg < prevy+threshLo) {     
      t = millis();
      b = 60000 / (t-prevt);
      if ((b >= 20) && (b <= 250) && (InPeak < 100/SamplePeriod)) { // peak must be narrow and BPM must be reasonable
        if (ignoreBeats > 0) {
          ignoreBeats--;
        } else {
          bpm = (bpm*7+b)/8; // smoothing
          bpm += constrain(b-bpm,-1,+1);
          
          tft.fillRect(0, 0, 70, 26, 0xffaa00);
          tft.setCursor(4,4);
          tft.setTextColor(ST77XX_WHITE);
          tft.setTextSize(3);
          tft.print(bpm);
            
//          centrePeak(t-prevt, x);
        }

      }
      prevt = t;      
      InPeak = 0;
    }    
  } else {
    if (ecg > prevy+threshHi) 
      InPeak = 1;
  }  
}
//-----------------------------------------------------------------------------
// DrawTraceSmall
//-----------------------------------------------------------------------------
void DrawTraceSmall(uint8_t y)
{
  static uint16_t x = 0;
  static uint8_t pt = 0;
  static uint8_t py = 0;
  static uint8_t yofs = 0;
  int i, x4, y4;

  x++;

  x = x % (TFT_WIDTH*2);
  x4 = x/2;
  y4 = y/2;
  
  if (x == 0) {
    yofs = (yofs+TFT_HEIGHT/2);
    if (yofs >= TFT_HEIGHT)
      yofs = 0;
  }

  if (x % 2 == 0) {
  if (x4*4*5 % (1000/SamplePeriod) == 0)
      tft.drawFastVLine(x4, yofs, TFT_HEIGHT/2, ST77XX_RED); else
      tft.drawFastVLine(x4, yofs, TFT_HEIGHT/2, ST77XX_BLACK);
  
  }

  y4 = constrain(TFT_HEIGHT/2-y4, 0,TFT_HEIGHT/2-1);
  if (y4 >= py)
    tft.drawFastVLine(x4, py+yofs, y4-py+1, ST77XX_WHITE); else
    tft.drawFastVLine(x4, y4+yofs, py-y4+1, ST77XX_WHITE);

  py = y4;

  calcBPM(y, x);
}
//-------------------------------------------------------------------------
// FilterLowPass
//   Low Pass Filter
//-------------------------------------------------------------------------
int FilterLowPass(int ecg)
{
    static int py = 0;
    static int ppy = 0;
    static int ppecg = 0;
    static int pecg = 0;
    int y;
    static int mid = 0;

    const long filt_a0 = 8775;
    const long filt_a1 = 17550;
    const long filt_a2 = 8775;
    const long filt_b1 = -50049;
    const long filt_b2 = 19612;

    if (ecg > mid) 
      mid++; else 
      mid--;
    
    ecg -= mid; // to remove DC offset
    y = (filt_a0*ecg + filt_a1*pecg + filt_a2*ppecg - filt_b1*py - filt_b2*ppy) >> 16;
    ppy = py;
    py = y;
    ppecg = pecg;
    pecg = ecg;
    return constrain(y + mid,0,1023);
}
//-------------------------------------------------------------------------
void setup() {
  pinMode(EKG_IN, INPUT);
  pinMode(BUTTON, INPUT_PULLUP);
  pinMode(TFT_BL, OUTPUT);
  analogReference(EXTERNAL);
  
  tft.init(240, 240);
  tft.fillScreen(ST77XX_BLACK);
  digitalWrite(TFT_BL, HIGH);

  if (!SD.begin(SDCARD)) {
    analogWrite(TFT_BL, 80);
    tft.setCursor(4,20);
    tft.setTextSize(3);
    tft.print("Fehler SD-Karte");
    delay (2000);
  }
}

void loop() {

  int ecg;
  static unsigned long nextTime = 0;  
  unsigned long t;  
  
  t = millis();
  
  if (t > nextTime) {
    if (t > nextTime+10)
      nextTime = t; else
      nextTime = nextTime + 5;

    ecg = analogRead(EKG_IN);
    ecg = FilterLowPass(ecg);

    ButtonState = digitalRead(BUTTON);
    if (ButtonState) {                         //Funktion WriteDisplay machen!?
      if (lastButtonState != ButtonState) {
        ekgDaten.close();
      }
      analogWrite(TFT_BL, 80);
      DrawTraceSmall(ecg / 4);
      lastButtonState = ButtonState;      //gehört in die obere if-Schleife?!
    }
    else {                                    //Funktion StartSdCard machen!?
      if (lastButtonState != ButtonState) {
        tft.fillScreen(ST77XX_BLACK);
        if (!SD.begin(SDCARD)) {
          analogWrite(TFT_BL, 80);
          tft.setCursor(4,20);
          tft.setTextSize(3);
          tft.print("Fehler SD-Karte");
          while (!digitalRead(BUTTON));   //solange auf SD gestellt, bleibt die Meldung
          goto weitermachen;              //sonst geht es wieder von vorne los
        }
        analogWrite(TFT_BL, 80);
        tft.setCursor(4,20);
        tft.setTextSize(3);
        tft.print("Schreibe auf SD-Karte");
        lastButtonState = ButtonState;
        
        char name[] = "ekg00.txt";
        for (uint8_t i = 1; i < 100; i++) {
          name[3] = i/10 + '0';
          name[4] = i%10 + '0';
          if (SD.exists(name)) continue;
          ekgDaten = SD.open((name), FILE_WRITE);
//          ekgDaten = SD.open((name), O_WRITE | O_CREAT);
          break;
        }
        delay(2000);
        tft.fillScreen(ST77XX_BLACK);
        digitalWrite(TFT_BL, LOW);
      }
      WriteSdCard(ecg);
    }
  weitermachen:;
  } 
}
